unit uGrid;

interface

uses
  System.SysUtils, System.Classes, JS, WEBLib.JSON, Web, WEBLib.Graphics, WEBLib.Controls,
  WEBLib.Forms, WEBLib.Dialogs, WEBLib.ExtCtrls, Vcl.Controls, WEBLib.TreeNodes,
  Vcl.ComCtrls, WEBLib.ComCtrls, WEBLib.ContinuousScroll, Vcl.StdCtrls,
  WEBLib.StdCtrls, jsdelphisystem, Data.DB, WEBLib.CDS, WEBLib.DBCtrls,
  WEBLib.DB, WEBLib.Grids, Vcl.Grids, uDetail, WEBLib.Buttons;

type
  TfrmGrid = class(TfrmDetail)
    mRowsJSON: TWebMemo;
    splMain: TWebSplitter;
    pGrid: TWebPanel;
    grdMain: TWebDBGrid;
    pNavigation: TWebPanel;
    navMain: TWebDBNavigator;
    pLog: TWebPanel;
    mLog: TWebMemo;
    pNav: TWebPanel;
    pLogToolbar: TWebPanel;
    cbScroll: TWebCheckBox;
    sbLogClear: TWebSpeedButton;
    sbSetup: TWebSpeedButton;
    cbCell: TWebCheckBox;
    cbForceEditors: TWebCheckBox;
    sbRefresh: TWebSpeedButton;
    sbChange: TWebSpeedButton;
    cbDetach: TWebCheckBox;
    sbRefreshN: TWebSpeedButton;
    edN: TWebEdit;
    cbCalc: TWebCheckBox;
    sbProblem1: TWebSpeedButton;
    procedure sbLogClearClick(Sender: TObject);
    procedure grdMainKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure sbSetupClick(Sender: TObject);
    procedure grdMainCanEditCell(Sender: TObject; ACol, ARow: Integer;
      var CanEdit: Boolean);
    procedure grdMainGetEditText(Sender: TObject; ACol, ARow: Integer;
      var Value: string);
    procedure grdMainSetEditText(Sender: TObject; ACol, ARow: Integer;
      const Value: string);
    procedure grdMainValidateEdit(Sender: TObject; ACol, ARow: Integer;
      var Value: string);
    procedure grdMainGetCellEditor(Sender: TObject; ACol, ARow: Integer;
      var AEditor: TGridCellEditor);
    procedure grdMainGetCellData(Sender: TObject; ACol, ARow: Integer;
      AField: TField; var AValue: string);
    procedure grdMainGetCellClass(Sender: TObject; ACol, ARow: Integer;
      AField: TField; AValue: string; var AClassName: string);
    procedure grdMainCheckClick(Sender: TObject; ACol, ARow: Integer;
      Checked: Boolean);
    procedure grdMainGetCellChildren(Sender: TObject; ACol, ARow: Integer;
      AField: TField; AValue: string; AElement: TJSHTMLElementRecord);
    procedure sbRefreshClick(Sender: TObject);
    procedure sbChangeClick(Sender: TObject);
    procedure sbRefreshNClick(Sender: TObject);
    procedure sbProblem1Click(Sender: TObject);
  private
    { Private declarations }
    cdMain: TWebClientDataset;
    procedure LogClear();
    procedure Log(AValue: String; AddInfo: Boolean = False);
    procedure DatasetBeforeOpen(DataSet: TDataSet);
    procedure DatasetAfterOpen(DataSet: TDataSet);
    procedure DatasetBeforeInsert(DataSet: TDataSet);
    procedure DatasetAfterInsert(DataSet: TDataSet);
    procedure DatasetOnNewRecord(DataSet: TDataSet);
    procedure DatasetBeforeEdit(DataSet: TDataSet);
    procedure DatasetAfterEdit(DataSet: TDataSet);
    procedure DatasetBeforeScroll(DataSet: TDataSet);
    procedure DatasetAfterScroll(DataSet: TDataSet);
    procedure DatasetBeforePost(DataSet: TDataSet);
    procedure DatasetAfterPost(DataSet: TDataSet);
    procedure DatasetBeforeDelete(DataSet: TDataSet);
    procedure DatasetAfterDelete(DataSet: TDataSet);
    procedure DatasetBeforeCancel(DataSet: TDataSet);
    procedure DatasetAfterCancel(DataSet: TDataSet);
    procedure DatasetOnCalcFields(DataSet: TDataSet);
    procedure FlagsOnChange(Sender: TField);
    procedure SimulateKey(AKey: String);
    function GetEditorName(AEditor: TGridCellEditor): String;
    procedure cdMaintsGetText(Sender: TField; var Text: string; DisplayText: Boolean);
    procedure cdMaindateGetText(Sender: TField; var Text: string; DisplayText: Boolean);
    procedure cdMaintimeGetText(Sender: TField; var Text: string; DisplayText: Boolean);
  public
    { Public declarations }
    procedure SetupForm(); override;
  end;

implementation

{$R *.dfm}

procedure TfrmGrid.LogClear();
begin
  mLog.Clear;
end;

procedure TfrmGrid.SetupForm();
var
  LFieldDefs: TFieldDefs;
  LFieldDef: TFieldDef;
  LColumn: TWebGridColumn;
begin
  inherited SetupForm();

  LogClear();

  cdMain := TWebClientDataSet.Create(nil);
  cdMain.BeforeOpen := DatasetBeforeOpen;
  cdMain.AfterOpen := DatasetAfterOpen;
  cdMain.BeforeInsert := DatasetBeforeInsert;
  cdMain.AfterInsert := DatasetAfterInsert;
  cdMain.OnNewRecord := DatasetOnNewRecord;
  cdMain.BeforeEdit := DatasetBeforeEdit;
  cdMain.AfterEdit := DatasetAfterEdit;
  cdMain.BeforeScroll := DatasetBeforeScroll;
  cdMain.AfterScroll := DatasetAfterScroll;
  cdMain.BeforePost := DatasetBeforePost;
  cdMain.AfterPost := DatasetAfterPost;
  cdMain.BeforeDelete := DatasetBeforeDelete;
  cdMain.AfterDelete := DatasetAfterDelete;
  cdMain.BeforeCancel := DatasetBeforeCancel;
  cdMain.AfterCancel := DatasetAfterCancel;
  cdMain.OnCalcFields := DatasetOnCalcFields;

  LFieldDefs := TFieldDefs.Create(nil);

  LFieldDef := LFieldDefs.AddFieldDef();
  LFieldDef.DataType := ftInteger;
  LFieldDef.DisplayName := 'Id';
  LFieldDef.Name := 'id';

  LFieldDef := LFieldDefs.AddFieldDef();
  LFieldDef.DataType := ftString;
  LFieldDef.DisplayName := 'Code';
  LFieldDef.Name := 'code';

  LFieldDef := LFieldDefs.AddFieldDef();
  LFieldDef.DataType := ftMemo;
  LFieldDef.DisplayName := 'Description';
  LFieldDef.Name := 'descr';

  LFieldDef := LFieldDefs.AddFieldDef();
  LFieldDef.DataType := ftFloat;
  LFieldDef.DisplayName := 'Amount';
  LFieldDef.Name := 'amount';

  LFieldDef := LFieldDefs.AddFieldDef();
  LFieldDef.DataType := ftDateTime;
  LFieldDef.DisplayName := 'Timestamp';
  LFieldDef.Name := 'ts';

  LFieldDef := LFieldDefs.AddFieldDef();
  LFieldDef.DataType := ftDate;
  LFieldDef.DisplayName := 'Date';
  LFieldDef.Name := 'date';

  LFieldDef := LFieldDefs.AddFieldDef();
  LFieldDef.DataType := ftTime;
  LFieldDef.DisplayName := 'Time';
  LFieldDef.Name := 'time';

  LFieldDef := LFieldDefs.AddFieldDef();
  LFieldDef.DataType := ftInteger;
  LFieldDef.DisplayName := 'Flags';
  LFieldDef.Name := 'flags';

  LFieldDef := LFieldDefs.AddFieldDef();
  LFieldDef.InternalCalcField := True;
  LFieldDef.DataType := ftBoolean;
  LFieldDef.DisplayName := '2';
  LFieldDef.Name := 'flags2';

  LFieldDef := LFieldDefs.AddFieldDef();
  LFieldDef.InternalCalcField := True;
  LFieldDef.DataType := ftBoolean;
  LFieldDef.DisplayName := '4';
  LFieldDef.Name := 'flags4';

  LFieldDef := LFieldDefs.AddFieldDef();
  LFieldDef.InternalCalcField := True;
  LFieldDef.DataType := ftBoolean;
  LFieldDef.DisplayName := '8';
  LFieldDef.Name := 'flags8';

  cdMain.FieldDefs := LFieldDefs;

  grdMain.Columns.Clear;

  LColumn := grdMain.Columns.Add;
  LColumn.ElementClassName := '';
  LColumn.DataField := 'id';
  LColumn.DataType := TColumnDataType.cdText;
  LColumn.Width := 50;
  LColumn.Title := 'Id';
  LColumn.TitleElementClassName := '';
  LColumn.Editor := TGridCellEditor.geRange;

  LColumn := grdMain.Columns.Add;
  LColumn.ElementClassName := '';
  LColumn.DataField := 'code';
  LColumn.DataType := TColumnDataType.cdText;
  LColumn.Width := 100;
  LColumn.Title := 'Code';
  LColumn.TitleElementClassName := '';
  LColumn.Editor := TGridCellEditor.geCombo;
  LColumn.ComboBoxItems.Clear;
  LColumn.ComboBoxItems.Add('1. code');
  LColumn.ComboBoxItems.Add('2. code');
  LColumn.ComboBoxItems.Add('3. code');

  LColumn := grdMain.Columns.Add;
  LColumn.ElementClassName := '';
  LColumn.DataField := 'descr';
  LColumn.DataType := TColumnDataType.cdText;
  LColumn.Width := 150;
  LColumn.Title := 'Description';
  LColumn.TitleElementClassName := '';
  LColumn.Editor := TGridCellEditor.geCombo;
  LColumn.ComboBoxItems.Clear;
  LColumn.ComboBoxItems.Add('1. description');
  LColumn.ComboBoxItems.Add('2. description');
  LColumn.ComboBoxItems.Add('3. description');

  LColumn := grdMain.Columns.Add;
  LColumn.ElementClassName := '';
  LColumn.DataField := 'amount';
  LColumn.DataType := TColumnDataType.cdText;
  LColumn.Width := 100;
  LColumn.Title := 'Amount';
  LColumn.TitleElementClassName := '';
  LColumn.Editor := TGridCellEditor.geNumber;

  LColumn := grdMain.Columns.Add;
  LColumn.ElementClassName := '';
  LColumn.DataField := 'ts';
  LColumn.DataType := TColumnDataType.cdText;
  LColumn.Width := 150;
  LColumn.Title := 'Timestamp';
  LColumn.TitleElementClassName := '';
  LColumn.Editor := TGridCellEditor.geText;

  LColumn := grdMain.Columns.Add;
  LColumn.ElementClassName := '';
  LColumn.DataField := 'date';
  LColumn.DataType := TColumnDataType.cdText;
  LColumn.Width := 100;
  LColumn.Title := 'Date';
  LColumn.TitleElementClassName := '';
  LColumn.Editor := TGridCellEditor.geDate;

  LColumn := grdMain.Columns.Add;
  LColumn.ElementClassName := '';
  LColumn.DataField := 'time';
  LColumn.DataType := TColumnDataType.cdText;
  LColumn.Width := 100;
  LColumn.Title := 'Time';
  LColumn.TitleElementClassName := '';
  LColumn.Editor := TGridCellEditor.geTime;

  LColumn := grdMain.Columns.Add;
  LColumn.ElementClassName := '';
  LColumn.DataField := 'flags';
  LColumn.DataType := TColumnDataType.cdText;
  LColumn.Width := 50;
  LColumn.Title := 'Flags';
  LColumn.TitleElementClassName := '';
  LColumn.Editor := TGridCellEditor.geText;

  LColumn := grdMain.Columns.Add;
  LColumn.ElementClassName := '';
  LColumn.DataField := 'flags2';
  LColumn.DataType := TColumnDataType.cdCheck;
  LColumn.Width := 50;
  LColumn.Title := '2';
  LColumn.TitleElementClassName := '';
  LColumn.Editor := TGridCellEditor.geText;

  LColumn := grdMain.Columns.Add;
  LColumn.ElementClassName := '';
  LColumn.DataField := 'flags4';
  LColumn.DataType := TColumnDataType.cdCheck;
  LColumn.Width := 50;
  LColumn.Title := '4';
  LColumn.TitleElementClassName := '';
  LColumn.Editor := TGridCellEditor.geText;

  LColumn := grdMain.Columns.Add;
  LColumn.ElementClassName := '';
  LColumn.DataField := 'flags8';
  LColumn.DataType := TColumnDataType.cdCheck;
  LColumn.Width := 50;
  LColumn.Title := '8';
  LColumn.TitleElementClassName := '';
  LColumn.Editor := TGridCellEditor.geText;

  Log('DataSource.DataSet := cdMain');
  DataSource.DataSet := cdMain;

  Log('grdMain.DataSource := DataSource');
  grdMain.DataSource := DataSource;

  Log('navMain.DataSource := DataSource');
  navMain.DataSource := DataSource;

  Log('cdMain.Rows');
  cdMain.Rows := TJSArray(TJSJSON.parseObject(mRowsJSON.Lines.Text));

  Log('cdMain.Open');
  cdMain.Open;

  cdMain.Fields.FindField('flags').ReadOnly := True;

  cdMain.Fields.FindField('ts').OnGetText := cdMaintsGetText;
  cdMain.Fields.FindField('date').OnGetText := cdMaindateGetText;
  cdMain.Fields.FindField('time').OnGetText := cdMaintimeGetText;

  cdMain.Fields.FindField('flags2').OnChange := FlagsOnChange;
  cdMain.Fields.FindField('flags4').OnChange := FlagsOnChange;
  cdMain.Fields.FindField('flags8').OnChange := FlagsOnChange;

  cdMain.Fields.FindField('flags2').Calculated := True;
  cdMain.Fields.FindField('flags4').Calculated := True;
  cdMain.Fields.FindField('flags8').Calculated := True;
end;

procedure TfrmGrid.DatasetOnCalcFields(DataSet: TDataSet);
var
  I: Integer;
  LField: TField;
  LFieldName: String;
  LBooleanField: TBooleanField;
  LValue: Boolean;
  Lflag: Integer;
  Lflags: Integer;
begin
  if cbCalc.Checked then
    Log('DatasetOnCalcFields');
  if DataSet=nil then
    Exit;
  Lflags := DataSet.FieldByName('flags').AsInteger;
  for I := 0 to DataSet.Fields.Count-1 do
  begin
    LField := DataSet.Fields[I];
    LFieldName := LField.FieldName;
    if LField.ClassType = TBooleanField then
    begin
      LBooleanField := TBooleanField(LField);
      Lflag := StrToIntDef(Copy(LFieldName, 6, Length(LFieldName)-5), 0);
      LValue := ((Lflags and Lflag) = Lflag);
      if (LBooleanField.AsBoolean <> LValue) then
        LBooleanField.AsBoolean := LValue;
    end;
  end;
end;

procedure TfrmGrid.FlagsOnChange(Sender: TField);
var
  LField: TField;
  LFieldName: String;
  Lflags: Integer;
  Lflags_new: Integer;
  Lflag: Integer;
begin
  Log('FlagsOnChange');
  if (Sender=nil) then
    Exit;
  LField := Sender.DataSet.FindField('flags');
  if LField=nil then
    Exit;
  LFieldName := Sender.FieldName;
  Lflags := LField.AsInteger;
  Lflag := StrToIntDef(Copy(LFieldName, 6, Length(LFieldName)-5), 0);
  if Sender.AsBoolean then
    Lflags_new := (Lflags or Lflag)
  else
    Lflags_new := (Lflags - (Lflags and Lflag));
  if (Lflags<>Lflags_new) then
  begin
    LField.AsInteger := Lflags_new;
    Log('flagsOnChange old: ' + IntToStr(Lflags) + ' new: ' + IntToStr(Lflags_new) );
  end;
end;

procedure TfrmGrid.cdMaintsGetText(Sender: TField; var Text: string; DisplayText: Boolean);
begin
  Log('cdMaintsGetText');
  Text := DateTimeToStr(Sender.AsDateTime);
end;

procedure TfrmGrid.cdMaindateGetText(Sender: TField; var Text: string; DisplayText: Boolean);
begin
  Log('cdMaindateGetText');
  Text := DateToStr(Sender.AsDateTime);
end;

procedure TfrmGrid.cdMaintimeGetText(Sender: TField; var Text: string; DisplayText: Boolean);
begin
  Log('cdMaintimeGetText');
  Text := TimeToStr(Sender.AsDateTime);
end;

procedure TfrmGrid.Log(AValue: String; AddInfo: Boolean = False);
var
  DataSet: TDataSet;
begin
  FormatSettings.LongTimeFormat := 'hh.nn.ss.zzz';
  mLog.Lines.Add(DateTimeToStr(Now) + ' ' + AValue);
  if (AddInfo) and (cdMain<>nil) and (cdMain.Active) then
  begin
    DataSet := TDataSet(cdMain);
    mLog.Lines.Add('   State: ' + IntToStr(Integer(DataSet.State)));
    mLog.Lines.Add('   RecordCount: ' + IntToStr(DataSet.RecordCount));
    mLog.Lines.Add('   RecNo: ' + IntToStr(DataSet.RecNo));
  end;
end;

procedure TfrmGrid.sbChangeClick(Sender: TObject);
begin
  inherited;
  if (cdMain.State<>dsInsert) and (cdMain.State<>dsEdit) then
    cdMain.Edit;
  if (cdMain.State=dsInsert) then
    cdMain.FieldByName('id').AsInteger := cdMain.RecordCount
  else
    cdMain.FieldByName('amount').AsFloat := cdMain.FieldByName('amount').AsFloat + 1.2;
end;

procedure TfrmGrid.sbLogClearClick(Sender: TObject);
begin
  inherited;
  mLog.Lines.Clear;
end;

procedure TfrmGrid.sbProblem1Click(Sender: TObject);
begin
  inherited;
  cdMain.Insert;
  sbChangeClick(nil);
end;

procedure TfrmGrid.sbRefreshClick(Sender: TObject);
begin
  inherited;
  Log('sbRefreshClick');
  if cbDetach.Checked then
    grdMain.DataSource := nil;
  cdMain.Close;
  cdMain.Rows := TJSArray(TJSJSON.parseObject(mRowsJSON.Lines.Text));
  cdMain.Open;
end;

procedure TfrmGrid.sbRefreshNClick(Sender: TObject);
var
  I: Integer;
begin
  inherited;
  for I := 1 to StrToIntDef(edN.Text,10) do
  begin
    sbRefreshClick(nil);
  end;
end;

procedure TfrmGrid.sbSetupClick(Sender: TObject);
begin
  inherited;
  SetupForm();
end;

procedure TfrmGrid.DatasetBeforeOpen(DataSet: TDataSet);
begin
  Log('DatasetBeforeOpen', True);
end;

procedure TfrmGrid.DatasetAfterOpen(DataSet: TDataSet);
begin
  Log('DatasetAfterOpen', True);
  if cbDetach.Checked then
    grdMain.DataSource := DataSource;
end;

procedure TfrmGrid.DatasetBeforeInsert(DataSet: TDataSet);
begin
  Log('DatasetBeforeInsert', True);
end;

procedure TfrmGrid.DatasetAfterInsert(DataSet: TDataSet);
begin
  Log('DatasetAfterInsert', True);
end;

procedure TfrmGrid.DatasetOnNewRecord(DataSet: TDataSet);
begin
  Log('DatasetOnNewRecord', True);
end;

procedure TfrmGrid.DatasetBeforeEdit(DataSet: TDataSet);
begin
  Log('DatasetBeforeEdit', True);
end;

procedure TfrmGrid.DatasetAfterEdit(DataSet: TDataSet);
begin
  Log('DatasetAfterEdit', True);
end;

procedure TfrmGrid.DatasetBeforeScroll(DataSet: TDataSet);
begin
  if cbScroll.Checked then
    Log('DatasetBeforeScroll', True);
end;

procedure TfrmGrid.DatasetAfterScroll(DataSet: TDataSet);
begin
  if cbScroll.Checked then
    Log('DatasetAfterScroll', True);
end;

procedure TfrmGrid.DatasetBeforePost(DataSet: TDataSet);
begin
  Log('DatasetBeforePost', True);
end;

procedure TfrmGrid.DatasetAfterPost(DataSet: TDataSet);
begin
  Log('DatasetAfterPost', True);
end;

procedure TfrmGrid.DatasetBeforeDelete(DataSet: TDataSet);
begin
  Log('DatasetBeforeDelete', True);
end;

procedure TfrmGrid.DatasetAfterDelete(DataSet: TDataSet);
begin
  Log('DatasetAfterDelete', True);
end;

procedure TfrmGrid.DatasetBeforeCancel(DataSet: TDataSet);
begin
  Log('DatasetBeforeCancel', True);
end;

procedure TfrmGrid.DatasetAfterCancel(DataSet: TDataSet);
begin
  Log('DatasetAfterCancel', True);
end;

procedure TfrmGrid.grdMainCanEditCell(Sender: TObject; ACol, ARow: Integer;
  var CanEdit: Boolean);
begin
  inherited;
  Log('CanEditCell ARow: ' + IntToStr(ARow) + ' ACol: ' + IntToStr(ACol));
end;

procedure TfrmGrid.grdMainCheckClick(Sender: TObject; ACol, ARow: Integer;
  Checked: Boolean);
begin
  inherited;
  Log('CheckClick ARow: ' + IntToStr(ARow) + ' ACol: ' + IntToStr(ACol) + ' Checked ' + BoolToStr(Checked, True));
end;

procedure TfrmGrid.grdMainGetCellChildren(Sender: TObject; ACol, ARow: Integer; AField: TField; AValue: string; AElement: TJSHTMLElementRecord);
begin
  inherited;
  if cbCell.Checked then
    Log('GetCellChildren ARow: ' + IntToStr(ARow) + ' ACol: ' + IntToStr(ACol)
     + ' AField.FieldName: ' + AField.FieldName + ' AValue: ' + AValue
    );
  if AnsiSameText(AField.FieldName,'code') then
  begin

  end else
  if AnsiSameText(AField.FieldName,'descr') then
  begin

  end;
end;

procedure TfrmGrid.grdMainGetCellClass(Sender: TObject; ACol, ARow: Integer;
  AField: TField; AValue: string; var AClassName: string);
begin
  inherited;
  if cbCell.Checked then
    Log('GetCellClass ARow: ' + IntToStr(ARow) + ' ACol: ' + IntToStr(ACol) + ' AField.FieldName: ' + AField.FieldName + ' AClassName: ' + AClassName);
end;

procedure TfrmGrid.grdMainGetCellData(Sender: TObject; ACol, ARow: Integer;
  AField: TField; var AValue: string);
begin
  inherited;
  if cbCell.Checked then
    Log('GetCellData ARow: ' + IntToStr(ARow) + ' ACol: ' + IntToStr(ACol) + ' Value: ' + AValue);
end;

function TfrmGrid.GetEditorName(AEditor: TGridCellEditor): String;
begin
  case AEditor of
    geText: Result := 'geText';
    geNumber: Result := 'geNumber';
    geDate: Result := 'geDate';
    geTime: Result := 'geTime';
    geRange: Result := 'geRange';
    geColor: Result := 'geColor';
    geWeek: Result := 'geWeek';
    geMonth: Result := 'geMonth';
    geURL: Result := 'geURL';
    geEmail: Result := 'geEmail';
    geTel: Result := 'geTel';
    geMask: Result := 'geMask';
    geCombo: Result := 'geCombo';
  else
    Result := '';
  end;
end;

procedure TfrmGrid.grdMainGetCellEditor(Sender: TObject; ACol, ARow: Integer; var AEditor: TGridCellEditor);
var
  LColumn: TWebGridColumn;
  LFieldName: String;
  LFieldIndex: Integer;
  LRecordIndex: Integer;
begin
  inherited;
  LFieldIndex := ACol - grdMain.FixedCols;
  LRecordIndex := ARow - grdMain.FixedRows;
  LColumn := grdMain.Columns.Items[LFieldIndex];
  LFieldName := LColumn.DataField;
  if cbForceEditors.Checked then
  begin
    if AnsiSameText(LFieldName,'id') then
      AEditor := TGridCellEditor.geRange;
    if AnsiSameText(LFieldName,'code') then
      AEditor := TGridCellEditor.geCombo;
    if AnsiSameText(LFieldName,'descr') then
      AEditor := TGridCellEditor.geCombo;
    if AnsiSameText(LFieldName,'amount') then
      AEditor := TGridCellEditor.geNumber;
    if AnsiSameText(LFieldName,'ts') then
      AEditor := TGridCellEditor.geText;
    if AnsiSameText(LFieldName,'date') then
      AEditor := TGridCellEditor.geDate;
    if AnsiSameText(LFieldName,'time') then
      AEditor := TGridCellEditor.geTime;
  end;
  Log('GetCellEditor ARow: ' + IntToStr(ARow) + ' ACol: ' + IntToStr(ACol) + ' AEditor: ' + GetEditorName(AEditor) + ' FieldName: ' + LFieldName + ' LRecordIndex: ' + IntToStr(LRecordIndex) + ' LFieldIndex: ' + IntToStr(LFieldIndex));
end;

procedure TfrmGrid.grdMainGetEditText(Sender: TObject; ACol, ARow: Integer;
  var Value: string);
begin
  inherited;
  Log('GetEditText ARow: ' + IntToStr(ARow) + ' ACol: ' + IntToStr(ACol) + ' Value: ' + Value);
end;

procedure TfrmGrid.grdMainKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  Log('grdMainKeyDown Key: ' + IntToStr(Key));
  inherited KeyDown(Key, Shift);
end;

procedure TfrmGrid.grdMainSetEditText(Sender: TObject; ACol, ARow: Integer;
  const Value: string);
begin
  inherited;
  Log('SetEditText ARow: ' + IntToStr(ARow) + ' ACol: ' + IntToStr(ACol) + ' Value: ' + Value);
end;

procedure TfrmGrid.grdMainValidateEdit(Sender: TObject; ACol, ARow: Integer;
  var Value: string);
begin
  inherited;
  Log('ValidateEdit ARow: ' + IntToStr(ARow) + ' ACol: ' + IntToStr(ACol) + ' Value: ' + Value);
end;

procedure TfrmGrid.SimulateKey(AKey: String);
begin

asm
var event_init = {};
event_init['key'] = AKey;
event_init['code'] = '';
event_init['location'] = 0;
event_init['ctrlKey'] = false;
event_init['shiftKey'] = false;
event_init['altKey'] = false;
event_init['metaKey'] = false;
event_init['repeat'] = false;
event_init['isComposing'] = false;
event_init['charCode'] = 0;
event_init['keyCode'] = 0;
event_init['which'] = 0;
var event = new KeyboardEvent('keydown', event_init);  //keydown, keyup, keypress
var grid = document.getElementById("DBGRIDNormalCellTable").parentElement.parentElement;
grid.dispatchEvent(event);
end;

end;

//SimulateKey('ArrowDown');
//SimulateKey('ArrowUp');

end.
